{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE PackageImports #-}
{-# LANGUAGE RecordWildCards #-}

{-# LANGUAGE NoImplicitPrelude #-}

module Datarekisteri.Backend where

import Relude

import "cryptonite" Crypto.Random (MonadRandom(..))
import qualified "base64" Data.ByteString.Base64 as B64

import Control.Monad.Except (catchError)
import Control.Monad.Logger (runStderrLoggingT)
import Data.Default (def)
import Data.Map (findWithDefault)
import Data.Text (toLower, breakOn, stripPrefix)
import Database.Persist.Postgresql (withPostgresqlConn, runSqlConn)
import Network.HTTP.Types.Status (status500, status401)
import Network.Mail.Mime (renderSendMailCustom, Address(..))
import Network.Wai (Application)
import Network.Wai.Handler.Warp (Port, run)
import Network.Wai.Middleware.Cors (CorsResourcePolicy(..), cors)
import Network.Wai.Middleware.Gzip (gzip)
import System.Directory (findExecutable)
import System.Process (callProcess)

import qualified Options.Applicative as O

import Options.Applicative hiding (Success, header)
import Web.Scotty.Trans hiding (readEither)

import Datarekisteri.Backend.Sql (MonadSql)
import Datarekisteri.Backend.Utils (checkPassword)

import qualified Datarekisteri.Backend.Sql as Sql

import Datarekisteri.Backend.API
import Datarekisteri.Backend.Types
import Datarekisteri.Core.Types

import Paths_datarekisteri_backend

main :: IO ()
main = do
    config <- readConfig >>= checkSendmailPath
    runMigrations (configDbUrl config)
    serverApp config >>= run 3100

readConfig :: IO Config
readConfig = do
    execParser $ info (configOpts <**> helper)
            (fullDesc <> progDesc "Serve a GraphQL API for datarekisteri"
            <> O.header "Backend API server for datarekisteri")

checkSendmailPath :: Config -> IO Config
checkSendmailPath config = do
    sendmailPath <- maybe (error "Sendmail command or file not found!") pure =<<
        findExecutable (configSendmail config)
    pure $ config {configSendmail = sendmailPath}

configOpts :: Parser Config
configOpts = Config
    <$> strOption (long "sendmail" <> short 'm' <> metavar "COMMAND" <> value "sendmail" <> help "Sendmail command")
    <*> (Address
        <$> optional (strOption (long "sender-name" <> short 'n' <> metavar "NAME" <> help "Display name for email address"))
        <*> strOption (long "address" <> short 'a' <> metavar "EMAIL" <> help "Email address to send mail"))
    <*> option auto (long "port" <> short 'p' <> metavar "PORT" <> value 3100 <> help "Port to listen on")
    <*> strOption (long "db-url" <> short 'u' <> metavar "URL" <> value "postgres:///datarekisteri-backend" <> help "Postgresql database url")

runMigrations :: Text -> IO ()
runMigrations dbUrl = do
    migrationsPath <- getDataFileName "db/migrations"
    callProcess "dbmate" ["--url", toString dbUrl, "--migrations-dir", migrationsPath, "up"]

serverApp :: Config -> IO Application
serverApp config = scottyAppT (runAPIM config) $ do
    middleware $ gzip def
    middleware $ cors $ const $ Just CorsResourcePolicy
        { corsOrigins = Nothing -- all
        , corsMethods = ["POST"]
        , corsRequestHeaders = ["Authorization"]
        , corsExposedHeaders = Nothing
        , corsMaxAge = Just (60*60*24)
        , corsVaryOrigin = False
        , corsRequireOrigin = False
        , corsIgnoreFailures = False
        }
    post "/" $ do
        maybeAuthorization <- fmap toText <$> header "Authorization"
        let maybeBasic = parseBasic maybeAuthorization
            maybeBearer = parseBearer maybeAuthorization
            auth = case maybeBasic of
                     Just _ -> authBasic maybeBasic
                     Nothing -> authBearer maybeBearer
        auth $ do
            setHeader "Content-Type" "text/json"
            body >>= lift . runApp coreApp >>= raw

data BasicAuth = BasicAuth { emailAddress :: Email, password :: Text } deriving (Show)

data BearerToken = BearerToken Text deriving (Show)

parseBearer :: Maybe Text -> Maybe BearerToken
parseBearer auth = do
    [authType, authData] <- words <$> auth
    guard $ toLower authType == "bearer"
    pure $ BearerToken authData

authBearer :: Maybe BearerToken -> ActionT LText APIM a -> ActionT LText APIM a
authBearer Nothing m = m
authBearer (Just (BearerToken bearer)) m = do
    let getUserPermissions = do
            Right DBToken {..} <- lift $ dbGetTokenBySecret bearer
            permissions' <- fromMaybe mempty . readPermission <$> lift dbTokenPermissions
            DBUser {..} <- lift dbTokenUser
            userID <- lift dbUserId
            pure (Just userID, permissions')
    (user, permissions) <- getUserPermissions `catchError` const (pure (Nothing, mempty))
    flip local m $ \state -> state
        { stateCurrentUser = user
        , statePermissions = permissions
        }

parseBasic :: Maybe Text -> Maybe BasicAuth
parseBasic txt = do
    [authType, authData] <- words <$> txt
    guard $ toLower authType == "basic"
    (email, password) <- rightToMaybe $
        breakOn' ":" . decodeUtf8 <$> B64.decodeBase64 (encodeUtf8 authData)
    emailAddress <- toEmail email
    pure $ BasicAuth {..}
        where breakOn' x xs = let (fst, snd) = breakOn x xs
                              in (fst, fromMaybe "" $ stripPrefix x snd)


authBasic :: Maybe BasicAuth -> ActionT LText APIM a -> ActionT LText APIM a
authBasic Nothing m = m
authBasic (Just basic) m = do
    DBUser {..} <- verifyBasic basic
    permissions <- readPermission <$> lift dbUserPermissions >>=
        fromMaybeFail status500 "Internal server error"
    userID <- lift dbUserId
    flip local m $ \state -> state
        { stateCurrentUser = Just userID
        , statePermissions = permissions
        }

verifyBasic :: BasicAuth -> ActionT LText APIM (DBUser APIM)
verifyBasic BasicAuth {..} = do
    maybeUser <- lift $ dbGetUserByEmail emailAddress
    let unauthorized = do
            setHeader "WWW-Authenticate" "Basic realm=\"GraphQL API\", Bearer realm=\"GraphQL API\""
            raiseStatus status401 "Wrong password or email"
    case maybeUser of
      Left _ -> unauthorized
      Right user@DBUser {..} -> do
          correctPassword <- checkPassword password <$> lift dbUserPasswordHash
          unless correctPassword unauthorized
          pure user

newtype APIM a = APIM (ReaderT RequestState IO a)
    deriving (Functor, Applicative, Monad, MonadIO, MonadReader RequestState)

data RequestState = RequestState
    { stateCurrentUser :: Maybe UserID
    , statePermissions :: Map Scope Permission
    , stateConfig :: Config
    }

data Config = Config
    { configSendmail :: FilePath
    , configEmailAddress :: Address
    , configPort :: Port
    , configDbUrl :: Text
    }

instance MonadTime APIM where
    currentTime = liftIO currentTime

instance MonadSql APIM where
    runQuery query = do
        dbUrl <- fmap encodeUtf8 $ asks $ configDbUrl . stateConfig
        liftIO $ runStderrLoggingT $ withPostgresqlConn dbUrl $ runSqlConn query

-- TODO: Catch database exceptions into Left values
instance MonadDB APIM where
    dbUpdateUser = Sql.dbUpdateUser
    dbAddUser = Sql.dbAddUser
    dbAcceptApplication = Sql.dbAcceptApplication
    dbRejectApplication = Sql.dbRejectApplication
    dbVerifyEmail = Sql.dbVerifyEmail
    dbAddToken = Sql.dbAddToken
    dbAddKey = Sql.dbAddKey
    dbGetUser = Sql.dbGetUser
    dbGetUserByEmail = Sql.dbGetUserByEmail
    dbGetUsers = Sql.dbGetUsers
    dbGetApplications = Sql.dbGetApplications
    dbGetEmailVerificationSecret = Sql.dbGetEmailVerificationSecret
    dbGetTokenBySecret = Sql.dbGetTokenBySecret

instance MonadEmail APIM where
    sendEmail email = do
        sendmailPath <- asks $ configSendmail . stateConfig
        liftIO $ renderSendMailCustom sendmailPath ["-t"] email
    fromAddress = asks $ configEmailAddress . stateConfig

instance MonadRequest APIM where
    currentUser = asks stateCurrentUser

instance MonadRandom APIM where
    getRandomBytes = liftIO . getRandomBytes

instance MonadPermissions APIM where
    currentPermissions = show <$> asks statePermissions
    defaultPermissions = pure $ show $ (fromList [(OwnProfile, ReadWrite)] :: Map Scope Permission)
    toPermissions = pure . fmap show . readPermission
    hasPermission scope permission = (>= permission) <$> findPermission scope
        where findPermission :: Scope -> APIM Permission
              findPermission scope@(Profile user) = selfPermissions scope user OwnProfile
              findPermission scope@(Tokens user) = selfPermissions scope user OwnTokens
              findPermission scope = findPermission' scope <$> asks statePermissions
              findPermission' :: Scope -> Map Scope Permission -> Permission
              findPermission' = findWithDefault None
              selfPermissions :: Scope -> UserID -> Scope -> APIM Permission
              selfPermissions scope user own = do
                  isSelf <- (Just user ==) <$> currentUser
                  let f = if isSelf then max <$> findPermission' own <*> findPermission' scope
                                    else findPermission' scope

                  f <$> asks statePermissions

runAPIM :: Config -> APIM a -> IO a
runAPIM config (APIM m) = runReaderT m RequestState
    { stateCurrentUser = Nothing
    , statePermissions = fromList []
    , stateConfig = config
    }

fromMaybeFail status err Nothing = raiseStatus status err
fromMaybeFail _ _ (Just x) = pure x
