{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE PackageImports #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}

{-# LANGUAGE NoImplicitPrelude #-}

module Datarekisteri.Backend.Types where

import Relude

import "cryptonite" Crypto.Random (MonadRandom(..))

import Control.Monad.Except (throwError)
import Data.Aeson (ToJSON(..), FromJSON(..))
import Data.ByteArray (ByteArray, ByteArrayAccess)
import Data.Morpheus.App.Internal.Resolving (Resolver, LiftOperation)
import Data.Morpheus.Types (MonadError, GQLError, GQLType)
import Data.Time (getCurrentTime)
import Database.Persist.Class (PersistField(..))
import Database.Persist.Sql (PersistFieldSql(..))
import Network.Mail.Mime (Mail, Address(..))

import Datarekisteri.Core.Types

forward :: Monad m => [a] -> m [Maybe a]
forward = pure . map Just

requirePermission :: (MonadPermissions m, MonadError GQLError m) => Scope -> Permission -> m ()
requirePermission scope permission = unlessM (hasPermission scope permission) $
    throwError $ "Insufficient permissions, " <> show permission <> " for "
        <> show scope <> " required."

data MemberData = MemberData
    { name :: Text
    , nickname :: Maybe Text
    , birthdate :: Date
    , homeplace :: Text
    , application :: Text
    , phoneNumber :: PhoneNumber
    } deriving (Show, Eq, Generic)

instance FromJSON MemberData
instance ToJSON MemberData

data ApplicationData = ApplicationData
    { email :: Email
    , phoneNumber :: PhoneNumber
    , password :: Text
    , name :: Text
    , nickname :: Maybe Text
    , birthdate :: Date
    , homeplace :: Text
    , application :: Text
    , registered :: Time
    , verificationSecret :: Text
    , passwordHash :: PasswordHash
    , permissions :: Text
    } deriving (Generic, Eq, Show)

data UpdateData = UpdateData
    { email :: Maybe Email
    , phoneNumber :: Maybe PhoneNumber
    , passwordHash :: Maybe PasswordHash
    , name :: Maybe Text
    , nickname :: Maybe Text
    , homeplace :: Maybe Text
    , application :: Maybe Text
    , user :: UserID
    , updateTime :: Time
    , verificationSecret :: Text
    } deriving (Generic, Eq, Show)

data NewKeyData = NewKeyData
    { comment :: Maybe Text
    , keyData :: ByteString
    , expires :: Maybe Time
    , uploaded :: Time
    , user :: UserID
    }
    deriving (Generic, Eq, Show)

newtype Cursor = Cursor Text
    deriving (Generic, Eq, Show)
    deriving anyclass GQLType

data Page a m = Page { pageData :: m a, cursor :: m (Maybe Cursor) }
    deriving (Generic, GQLType)

data NewTokenData = NewTokenData
    { comment :: Maybe Text
    , name :: Maybe Text
    , permissions :: Text 
    , tokenData :: Text
    , issued :: Time
    , expires :: Maybe Time
    , user :: UserID
    }
    deriving (Generic, Eq, Show)

newtype PasswordHash = PasswordHash ByteString
    deriving newtype (Eq, Show, Ord, Semigroup, Monoid, ByteArrayAccess,
        ByteArray, PersistField, PersistFieldSql)

data DBUser m = DBUser
    { dbUserId :: m UserID
    , dbUserEmail :: m (Maybe Email)
    , dbUserPendingEmail :: m (Maybe Email)
    , dbUserName :: m Text
    , dbUserNickname :: m Text
    , dbUserPhoneNumber :: m PhoneNumber
    , dbUserBirthdate :: m Date
    , dbUserHomeplace :: m Text
    , dbUserRegistered :: m Time
    , dbUserAccepted :: m (Maybe Time)
    , dbUserPermissions :: m Text
    , dbUserApplication :: m Text
    , dbUserPasswordHash :: m PasswordHash
    , dbUserTokens :: m [DBToken m]
    , dbUserKeys :: m [DBPGPKey m]
    , dbUserPrimaryKey :: m (Maybe (DBPGPKey m))
    }

data DBPGPKey m = DBPGPKey
    { dbPGPKeyId :: m KeyID
    , dbPGPKeyData :: m Base64
    , dbPGPKeyExpires :: m (Maybe Time)
    , dbPGPKeyUploaded :: m Time
    , dbPGPKeyComment :: m Text
    }

data DBToken m = DBToken
    { dbTokenId :: m TokenID
    , dbTokenUser :: m (DBUser m)
    , dbTokenName :: m (Maybe Text)
    , dbTokenData :: m Text
    , dbTokenComment :: m Text
    , dbTokenIssued :: m Time
    , dbTokenExpires :: m (Maybe Time)
    , dbTokenPermissions :: m Text
    }

class Monad m => MonadTime m where
    currentTime :: m Time

instance MonadTime IO where
    currentTime = Time <$> getCurrentTime

type DBEither a = Either String a

class Monad m => MonadDB m where
    dbUpdateUser :: UpdateData -> m (DBEither (DBUser m))
    dbAddUser :: ApplicationData -> m (DBEither (DBUser m))
    dbAcceptApplication :: UserID -> Time -> m (DBEither ())
    dbRejectApplication :: UserID -> m (DBEither ())
    dbVerifyEmail :: Text -> m (DBEither ())
    dbAddToken :: NewTokenData -> m (DBEither (DBToken m))
    dbAddKey :: NewKeyData -> m (DBEither (DBPGPKey m))
    dbGetUser :: UserID -> m (DBEither (DBUser m))
    dbGetUserByEmail :: Email -> m (DBEither (DBUser m)) -- XXX should this be Maybe instead
    dbGetUsers :: m [DBUser m]
    dbGetApplications :: m [DBUser m]
    dbGetTokenBySecret :: Text -> m (DBEither (DBToken m))
    dbGetEmailVerificationSecret :: UserID -> m (DBEither (Maybe Text))

class Monad m => MonadEmail m where
    sendEmail :: Mail -> m ()
    fromAddress :: m Address

class Monad m => MonadRequest m where
    currentUser :: m (Maybe UserID)

class Monad m => MonadPermissions m where
    hasPermission :: Scope -> Permission -> m Bool
    currentPermissions :: m Text
    defaultPermissions :: m Text
    toPermissions :: Text -> m (Maybe Text)

instance (MonadDB m, LiftOperation o) => MonadDB (Resolver o () m) where
    dbUpdateUser = fmap (fmap liftUser) . lift . dbUpdateUser
    dbAddUser = fmap (fmap liftUser) . lift . dbAddUser
    dbAcceptApplication user = lift . dbAcceptApplication user
    dbRejectApplication = lift . dbRejectApplication
    dbVerifyEmail = lift . dbVerifyEmail
    dbAddToken = fmap (fmap liftToken) . lift . dbAddToken
    dbAddKey = fmap (fmap liftKey) . lift . dbAddKey
    dbGetUser = fmap (fmap liftUser) . lift . dbGetUser
    dbGetUserByEmail = fmap (fmap liftUser) . lift . dbGetUserByEmail
    dbGetUsers = fmap (map liftUser) $ lift $ dbGetUsers
    dbGetApplications = fmap (map liftUser) $ lift $ dbGetApplications
    dbGetEmailVerificationSecret = lift . dbGetEmailVerificationSecret
    dbGetTokenBySecret = fmap (fmap liftToken) . lift . dbGetTokenBySecret

instance (MonadRequest m, LiftOperation o) => MonadRequest (Resolver o () m) where
    currentUser = lift currentUser

instance (MonadPermissions m, LiftOperation o) => MonadPermissions (Resolver o () m) where
    hasPermission scope permission = lift $ hasPermission scope permission
    defaultPermissions = lift defaultPermissions
    currentPermissions = lift currentPermissions
    toPermissions = lift . toPermissions

instance (MonadEmail m, LiftOperation o) => MonadEmail (Resolver o () m) where
    sendEmail = lift . sendEmail
    fromAddress = lift fromAddress

instance (MonadRandom m, LiftOperation o) => MonadRandom (Resolver o () m) where
    getRandomBytes = lift . getRandomBytes

instance (MonadTime m, LiftOperation o) => MonadTime (Resolver o () m) where
    currentTime = lift currentTime

liftUser :: (MonadTrans t, Monad m, Monad (t m)) => DBUser m -> DBUser (t m)
liftUser DBUser {..} = DBUser
    { dbUserId = lift dbUserId
    , dbUserEmail = lift dbUserEmail
    , dbUserPendingEmail = lift dbUserPendingEmail
    , dbUserName = lift dbUserName
    , dbUserNickname = lift dbUserNickname
    , dbUserPhoneNumber = lift dbUserPhoneNumber
    , dbUserBirthdate = lift dbUserBirthdate
    , dbUserHomeplace = lift dbUserHomeplace
    , dbUserRegistered = lift dbUserRegistered
    , dbUserAccepted = lift dbUserAccepted
    , dbUserPermissions = lift dbUserPermissions
    , dbUserApplication = lift dbUserApplication
    , dbUserPasswordHash = lift dbUserPasswordHash
    , dbUserTokens = map liftToken <$> lift dbUserTokens
    , dbUserKeys = map liftKey <$> lift dbUserKeys
    , dbUserPrimaryKey = fmap liftKey <$> lift dbUserPrimaryKey
    }

liftToken :: (MonadTrans t, Monad m, Monad (t m)) => DBToken m -> DBToken (t m)
liftToken DBToken {..} = DBToken
    { dbTokenId = lift dbTokenId
    , dbTokenUser = fmap liftUser $ lift dbTokenUser
    , dbTokenName = lift dbTokenName
    , dbTokenData = lift dbTokenData
    , dbTokenComment = lift dbTokenComment
    , dbTokenIssued = lift dbTokenIssued
    , dbTokenExpires = lift dbTokenExpires
    , dbTokenPermissions = lift dbTokenPermissions
    }

liftKey :: (MonadTrans t, Monad m, Monad (t m)) => DBPGPKey m -> DBPGPKey (t m)
liftKey DBPGPKey {..} = DBPGPKey
    { dbPGPKeyId = lift dbPGPKeyId
    , dbPGPKeyData = lift dbPGPKeyData
    , dbPGPKeyExpires = lift dbPGPKeyExpires
    , dbPGPKeyUploaded = lift dbPGPKeyUploaded
    , dbPGPKeyComment = lift dbPGPKeyComment
    }
